import logging

from aiohttp import ClientResponse

from ...utils.config import CONFIG
from ..exceptions.validator import (
    EventValidationError
)
from ..exceptions.vkapi import (
    VkApiError,
    AuthError,
    LongPollError
)
from ...utils.validation_reports import make_it_colored as color

logger = logging.getLogger('event-validator')

class EventValidator:

    _requiredFields = ['response', 'error']
    _pollingRequiredFields = ['updates', 'failed']
    _errorFields = [
        _requiredFields[-1],
        _pollingRequiredFields[-1]
    ]

    async def _isJsonValid(response: ClientResponse) -> bool:
        '''
        Check the JSON in the response
        '''
        logger.debug(f"Event Validaion: Checking Content-Type header for JSON...")

        _content_type = response.headers.get('Content-Type', None)

        if not _content_type or _content_type != 'application/json':
            logger.error(f"Response doesn't have JSON content")
            return False
        
        return True
    
    async def _fieldsAreValid(content: dict,
                              fields: dict={}) -> bool:
        '''
        Check required fields in the response
        '''
        logger.debug(f"Event Validation: Checking for one of the required response fields...")

        for field in fields:
            if field in content:
                return True
        else:
            logger.error(f"Response doesn't contain any of the requirement fields")
            return False
    
    async def _haveErrors(content: dict) -> bool:
        '''
        Check for errors in the response
        '''
        logger.debug(f"Event Validation: Checking for errors in the response...")

        for field in EventValidator._errorFields:
            if field in content:
                if field == "error":
                    error_code = content[field]["error_code"]
                    error_msg = content[field]["error_msg"]
                    request_params = content[field]["request_params"]
                    
                    if content[field]["error_code"] in [5, 1116]:
                        raise AuthError(error_code, error_msg, request_params)
                    
                    raise VkApiError(error_code, error_msg, request_params)
                
                if field == "failed":
                    failed_code = content[field]
                    raise LongPollError(failed_code, content["ts"] if "ts" in content else None)
                
                return True
            
        else:
            return False

    async def validate(response: ClientResponse,
                       from_polling: bool = False) -> dict:
        '''
        Validating response from API
        '''
        logger.debug(f"Validating event response for [{response.method}] \"{response.url}\"...")

        jsonIsValid = await EventValidator._isJsonValid(response)
        fieldsAreValid = "-"
        haveErrors = "-"
        content = None

        if jsonIsValid:
            content = await response.json()
            
            fieldsAreValid = await EventValidator._fieldsAreValid(content, 
                                                                  EventValidator._requiredFields if not from_polling else EventValidator._pollingRequiredFields)
            haveErrors = await EventValidator._haveErrors(content)

            if fieldsAreValid and not haveErrors:
                logger.debug("Event response is valid")
                return content
        
        logger.warning(f"Event response is not valid")
        logger.warning(f'{"JSON is valid":<15}|{"":>1}{"Fields are valid":<18}|{"":>1}{"Has no errors":<15}|{"":>1}{"content":<15}')
        logger.warning(f'{color(jsonIsValid, indent=15)}|{"":>1}{color(fieldsAreValid, indent=18)}|{"":>1}{color(not haveErrors, indent=15)}|{"":>1}{color(content)}')

        raise EventValidationError(jsonIsValid, fieldsAreValid, content)


import pytest
import pytest_asyncio

from twilight_vk.http.async_http import Http

@pytest.fixture
def http_params():
    return {
        "test_key": "123",
        "test_str": "abc123"
    }

@pytest.fixture
def http_data():
    return {
        "test_key": "test_value",
        "test_str": "abcd"
    }

@pytest_asyncio.fixture
async def http_client():
    httpClient = Http(headers={"X-Test": "test"}, timeout=10)
    yield httpClient
    await httpClient.close()


from .base import BaseMethodsGroup

class Groups(BaseMethodsGroup):

    async def getById(self,
                      group_ids: list[str] | str = None,
                      group_id: str = None,
                      fields: str = None) -> dict:
        '''
        Возвращает информацию о заданном сообществе или о нескольких сообществах.
        (см. https://dev.vk.com/ru/method/groups.getById)
        '''
        values = {
            "group_ids": group_ids,
            "group_id": abs(group_id) if group_id is not None else self.__group_id__,
            "fields": fields,
            "v": self.__api_version__
        }
        response = await self.base_api.base_get_method(api_method=f"{self.method}.getById",
                                                       values=values)
        return response
    
    async def getLongPollServer(self,
                                group_id:int=None) -> dict:
        
        '''
        Возвращает данные для подключения к Bots Longpoll API.
        (см. https://dev.vk.com/ru/method/groups.getLongPollServer)
        '''

        values = {
            "group_id": abs(group_id) if group_id is not None else self.__group_id__,
            "v": self.__api_version__
        }
        response = await self.base_api.base_get_method(api_method=f"{self.method}.getLongPollServer",
                                                       values=values)
        return response
    


import json
import logging
from typing import Any

from ...utils.keyboard.keyboard import KeyboardMarkup

logger = logging.getLogger('reqst-validator')

class RequestValidator:

    async def _getValidValue(value: 'Any') -> 'Any':
        '''
        Преобразует значение Python в корректное значение для запроса в API
        '''
        logger.debug(f"Validating value: {value} with type: {type(value)}")
        if isinstance(value, bool):
            return "true" if value == True else "false"
        
        if isinstance(value, dict):
            return json.dumps(value)

        if isinstance(value, list):
            return ",".join([f"{val}" for val in value])
        
        if isinstance(value, str):
            return f"{value}"
        
        if isinstance(value, int) or isinstance(value, float):
            return value
        
        if isinstance(value, KeyboardMarkup):
            return json.dumps(value.getMarkup())

        return None

    async def validate(values: dict) -> dict:
        _validValues = {}

        logger.debug(f"Validating request parameters...")
        logger.debug(f"Parameters to validate: {values}")

        for key, value in values.items():
            if value not in ['', None]:
                _validValues.setdefault(key, await RequestValidator._getValidValue(value))

        logger.debug(f"Parameters has been validated. Result: {_validValues}")

        return _validValues


import pytest
import pytest_asyncio

from twilight_vk.utils.keyboard import (
    KeyboardMarkup,
    TextActionKeyboardButton,
    LocationKeyboardButton,
    VkPayKeyboardButton
)
from twilight_vk.utils.types.keyboard_colors import KeyboardColor
from tests.fixtures.keyboard import (
    keyboard_markup,
    inline_keyboard_markup
)

@pytest.mark.asyncio
async def test_keyboard(keyboard_markup: dict, inline_keyboard_markup: dict):
    _inline_keyboard = KeyboardMarkup(inline=True,
                                      buttons=[
                                          [
                                              TextActionKeyboardButton(
                                                  label="TestButton",
                                                  color=KeyboardColor.POSITIVE
                                              )
                                          ]
                                      ])
    assert _inline_keyboard.getMarkup() == inline_keyboard_markup

    _keyboard = KeyboardMarkup(one_time=True,
                               inline=False,
                               buttons=[
                                   [
                                       TextActionKeyboardButton(
                                           label="TestButton",
                                           payload="{\"button\": \"1\"}",
                                           color=KeyboardColor.SECONDARY
                                       ),
                                       LocationKeyboardButton(
                                           payload="{\"button\": \"2\"}"
                                       )
                                   ],
                                   [
                                       VkPayKeyboardButton(
                                           hash="action=transfer-to-group&group_id=12345&aid=10"
                                       )
                                   ]
                               ])
    assert _keyboard.getMarkup() == keyboard_markup


import asyncio

import pytest
import pytest_asyncio

from twilight_vk import TwilightVK
from twilight_vk.framework.rules import TrueRule
from twilight_vk.utils.types.event_types import BotEventType
from tests.fixtures.handlers import (
    bot,
    response
)

@pytest.mark.asyncio
async def test_labeler(bot: TwilightVK, caplog):
    
    @bot.on_event.all(TrueRule())
    async def foo(event: dict):
        return "OK"
    
    for handler_name in bot.__bot__._router._handlers.keys():
        assert bot.__bot__._router._handlers[handler_name]._funcs[0].get("func", False) == foo
        assert isinstance(bot.__bot__._router._handlers[handler_name]._funcs[0].get("rules")[0], TrueRule)

@pytest.mark.asyncio
async def test_event_handlers(bot: TwilightVK, caplog, monkeypatch, response: dict):
    
    @bot.on_event.message_new(TrueRule())
    async def handle_message(event: dict):
        assert isinstance(event, dict)
        assert event.keys() == {"type": "test", "object": {}}.keys()
        assert event["object"].keys() == {"client_info": {}, "message": {}}.keys()
        bot.logger.info(f"TEST_MESSAGE_NEW")
    
    @bot.on_event.raw(BotEventType.LIKE_ADD)
    async def handle_like(event: dict):
        assert isinstance(event, dict)
        assert event.keys() == {"type": "test", "object": {}}.keys()
        assert event["object"].keys() == {
            "liker_id": 123,
            "object_type": "post",
            "object_owner_id": 123,
            "object_id": 1,
            "thread_reply_id": 1,
            "post_id": 1
        }.keys()
        bot.logger.info(f"TEST_LIKE_ADD")

    async def fake_messageSend(*args, **kwargs):
        return True

    for handler_name in bot.__bot__._router._handlers.keys():
        monkeypatch.setattr(bot.__bot__._router._handlers[handler_name].vk_methods.messages, "send", fake_messageSend)

    results = await asyncio.gather(
        *(bot.__bot__._router.route(event) for event in response["updates"]),
        return_exceptions=False
    )

    for result in results:

        if result is None:
            assert True
            continue

        assert result
    
    assert "TEST_MESSAGE_NEW" in caplog.text
    assert "TEST_LIKE_ADD" in caplog.text


import asyncio

import pytest
import pytest_asyncio
import pytest_mock

from twilight_vk import TwilightVK
from twilight_vk.utils.types.twi_states import TwiVKStates
from tests.fixtures.polling import (
    bot,
    MockEventGenerator
)

@pytest.mark.asyncio
async def test_authorization(bot: TwilightVK, caplog, monkeypatch, mocker):

    http_validate_mock = mocker.AsyncMock()
    http_validate_mock.side_effect = lambda response, *args, **kwargs: response
    mocker.patch('twilight_vk.framework.methods.base.HttpValidator.validate', http_validate_mock)

    monkeypatch.setattr(bot.__bot__.vk_methods.groups.base_api.httpClient, "get", MockEventGenerator("auth fail").get)
    await bot.__bot__.auth()
    assert bot.__bot__.__server__ is None
    assert bot.__bot__.__key__ is None
    assert bot.__bot__.__ts__ is None
    assert "Authrization error: [5] User authorization failed: invalid access_token (4)." in caplog.text

    monkeypatch.setattr(bot.__bot__.vk_methods.groups.base_api.httpClient, "get", MockEventGenerator("auth success").get)
    await bot.__bot__.auth()
    assert bot.__bot__.__server__ is not None
    assert bot.__bot__.__key__ is not None
    assert bot.__bot__.__ts__ is not None
    assert "Authorized" in caplog.text

@pytest.mark.asyncio
async def test_check_event(bot: TwilightVK, caplog, monkeypatch, mocker):
    
    http_validate_mock = mocker.AsyncMock()
    http_validate_mock.side_effect = lambda response, *args, **kwargs: response
    mocker.patch('twilight_vk.framework.polling.bots_longpoll.HttpValidator.validate', http_validate_mock)

    monkeypatch.setattr(bot.__bot__.httpClient, "get", MockEventGenerator("message_typing_state").get)
    assert await bot.__bot__.check_event() == await MockEventGenerator("message_typing_state").json()

@pytest.mark.asyncio
async def test_polling_listen(bot: TwilightVK, caplog, monkeypatch, mocker):
    
    http_validate_mock = mocker.AsyncMock()
    http_validate_mock.side_effect = lambda response, *args, **kwargs: response
    mocker.patch('twilight_vk.framework.polling.bots_longpoll.HttpValidator.validate', http_validate_mock)

    monkeypatch.setattr(bot.__bot__.httpClient, "get", MockEventGenerator("message_new").get)
    async for event in bot.__bot__.listen():
        assert event == await MockEventGenerator("message_new").json()
        break

@pytest.mark.asyncio
async def test_failed_polling(bot: TwilightVK, caplog, monkeypatch, mocker):
    
    http_validate_mock = mocker.AsyncMock()
    http_validate_mock.side_effect = lambda response, *args, **kwargs: response
    mocker.patch('twilight_vk.framework.polling.bots_longpoll.HttpValidator.validate', http_validate_mock)

    bot._state = TwiVKStates.READY

    bot.should_stop()

    monkeypatch.setattr(bot.__bot__.vk_methods.groups.base_api.httpClient, "get", MockEventGenerator("auth success").get)

    failed_codes = ["1", "2", "3"]

    for code in failed_codes:
        monkeypatch.setattr(bot.__bot__.httpClient, "get", MockEventGenerator(f"failed event {code}").get)
        await bot.__bot__.check_event()

        if code == "1":
            assert bot.__bot__.__ts__ == 321
    
    assert "The event history is outdated or has been partially lost. The application can receive events further using the new \"ts\" value from the response." in caplog.text
    assert "The key is expired. Getting new one..." in caplog.text
    assert "The information is lost. Reauthorizing..." in caplog.text

@pytest.mark.asyncio
async def test_graceful_stop_polling(bot: TwilightVK, caplog, monkeypatch):

    bot._state = TwiVKStates.READY
    
    bot.should_stop()
    assert "Polling will be stopped as soon as the current request will be done. Please wait" in caplog.text
    assert bot._state == TwiVKStates.SHUTTING_DOWN


import asyncio
from typing import Any

import pytest
import pytest_asyncio

from twilight_vk.utils.event_loop import TwiTaskManager
from twilight_vk.utils.twiml import TwiML
from tests.fixtures.utils import (
    loop,
    log_fixture,
    twiml,
    parse_dataset,
    mentions_dataset
)

async def fake_coro1(wrapper: TwiTaskManager):
    wrapper.logger.info("Hello")

async def fake_coro2(wrapper: TwiTaskManager):
    wrapper.logger.info("World")

async def fake_coro3(wrapper: TwiTaskManager):
    wrapper.logger.info("fake-coro3 start")
    await wrapper._handle_sigterm()
    await asyncio.sleep(5)
    wrapper.logger.info("fake-coro3 end")

async def fake_coro4(wrapper: TwiTaskManager):
    wrapper.logger.info("Test1")

async def fake_coro5(wrapper: TwiTaskManager):
    wrapper.logger.info("Test2")


def test_twi_task_manager(loop: TwiTaskManager,
                          log_fixture: Any):
    loop.add_task(fake_coro1(loop))
    loop.add_task(fake_coro2(loop))
    loop.run()
    assert "Hello" in log_fixture.text
    assert "World" in log_fixture.text

    loop.add_task(fake_coro3(loop))
    loop.run()
    assert "fake-coro3 start" in log_fixture.text
    assert "fake-coro3 end" not in log_fixture.text

    loop.add_task([fake_coro4(loop), fake_coro5(loop)])
    loop.run()
    assert "Test1" in log_fixture.text
    assert "Test2" in log_fixture.text

    loop.stop()
    assert "Stopping all tasks..." in log_fixture.text
    assert "Tasks was stopped" in log_fixture.text

@pytest.mark.asyncio
async def test_twiml_parse_params(twiml: TwiML,
                                  parse_dataset: dict):
    for _temp in range(len(parse_dataset.get("templates"))):
        twiml.update_template(parse_dataset.get("templates")[_temp])
        for _msg in range(len(parse_dataset.get("messages"))):
            result = await twiml.parse(parse_dataset.get("messages")[_msg])
            assert result == parse_dataset.get("results")[_temp][_msg]

@pytest.mark.asyncio
async def test_twiml_extract_mentions(twiml: TwiML,
                                      mentions_dataset: dict):
    for _msg in range(len(mentions_dataset.get("messages"))):
        mentions = await twiml.extract_mentions(mentions_dataset.get("messages")[_msg])
        assert mentions == mentions_dataset.get("results")[_msg]


from typing import TYPE_CHECKING
from typing import List, Callable
import logging

import asyncio

from ..rules import *
from ...utils.types.response import Response
from ..validators import ResponseValidator
from ..exceptions.handler import (
    ResponseHandlerError
)

if TYPE_CHECKING:
    from ..methods import VkMethods

class BASE_EVENT_HANDLER:

    def __init__(self,
                 vk_methods:'VkMethods',):
        '''
        BASE_EVENT_HANDLER is base handler for vk_api events.
        It contsains the base logic for event handler.
        All child event handlers must inherit it

        :param vk_methods: Initialized VkMethods class which allows to use VK API methods
        :type vk_methods: VkMethods
        '''
        self.logger = logging.getLogger(f"event-handler")
        self.logger.log(1, f"{self.__class__.__name__} event handler is initiated")

        self.vk_methods = vk_methods

        self._funcs: List[dict[list[BaseRule], Callable]] = []
    
    def add(self, 
                func, 
                rules: list):
        '''
        Allows to add callable functions into this handler
        '''
        self._funcs.append(
            {
                "rules": rules,
                "func": func
            }
        )
        self.logger.log(1, f"{func.__name__}() was added to {self.__class__.__name__} "\
                          f"with rules: {[f"{rule.__class__.__name__}" for rule in rules]}")
    
    async def _checkRule(self,
                            rule: BaseRule, 
                            event: dict):
        '''
        Checking rule result for current function and event
        '''
        if not hasattr(rule, "methods") or rule.methods is None:
            await rule._linkVkMethods(self.vk_methods)

        result = await rule._check(event)

        return result

    async def _checkRules(self, func, handler, event):
        '''
        Checks all results
        '''
        self.logger.debug(f"Checking rules for {func.__name__} from {self.__class__.__name__}...")
        rule_results = await asyncio.gather(
            *(self._checkRule(rule, event) for rule in handler["rules"]),
            return_exceptions=False
        )
        self.logger.debug(f"Rules check results: {rule_results}")
        self.logger.debug(f"{func.__name__}'s rules was checked")

        should_stop = False

        for rule in rule_results:
            if isinstance(rule, Exception):
                self.logger.error(f"Got an exception in rules check results. [{rule.__class__.__name__}({rule})]")
                should_stop = True
            if rule is False:
                self.logger.debug(f"One of the rules has returned False")
                should_stop = True

        if should_stop:
            return False
        
        return rule_results
    
    async def _extractArgs(self, rule_results:list):
        '''
        Extracts all args from rule_results after regex rules
        '''
        args = {}
        if isinstance(rule_results, list):
            for result in rule_results:
                if isinstance(result, dict):
                    for key, value in result.items():
                        args.setdefault(key, value)
        return args
   
    async def _handleOutput(self,
                               func,
                               callback: 'str | Response',
                               event: dict):
        '''
        Handles the output from functions which was added to handler
        '''    
        self.logger.debug(f"Handling response for {self.__class__.__name__}.{func.__name__}...")

        if callback is None:
            self.logger.debug(f"Callback is None. Output handling was skipped")
            return True
        
        callback = await ResponseValidator.validate(callback,
                                                    event)
        
        if isinstance(callback, Response):
            response = await self.vk_methods.messages.send(**callback.getData())
            return response
        
        raise ResponseHandlerError(callback, isinstance(callback, Response | None))
        
    async def _callFunc(self, handler:dict, event:dict):
        '''
        Executes separate function from self.__funcs__ list
        '''
        func = handler["func"]

        rule_results = await self._checkRules(func, handler, event)

        extracted_args = await self._extractArgs(rule_results)
        
        if rule_results is not False:
            self.logger.debug(f"Calling {func.__name__} from {self.__class__.__name__}...")
            if asyncio.iscoroutinefunction(func):
                response = await func(event, **extracted_args)
            else:
                response = func(event, **extracted_args)

            result = await self._handleOutput(func, response, event)
            self.logger.debug(f"{func.__name__} from {self.__class__.__name__} was called")

    async def execute(self, 
                      event: dict, 
                      in_gather: bool = True):
        '''
        Handles the event

        :param in_gather: Flag which switches the handling mode (one after one or all in one)
        :type in_gather: bool
        '''
        self.logger.debug(f"{self.__class__.__name__} is working right now...")
        self.logger.debug(f"Handling mode: {"all in one" if in_gather else "one after one"}")

        if in_gather:
            await asyncio.gather(
                *(self._callFunc(handler, event) for handler in self._funcs),
                return_exceptions=False
            )
        else:
            for handler in self._funcs:
                await self._callFunc(handler, event)

        self.logger.debug(f"{self.__class__.__name__} is finished handling the event")


import logging

from .base import KeyboardButton
from ...types.keyboard_colors import KeyboardColor

logger = logging.getLogger('keyboard')

class TextActionKeyboardButton(KeyboardButton):
    '''
    Кнопка, которая отправляет сообщение с текстом в Label
    (см. https://dev.vk.ru/ru/api/bots/development/keyboard#text)
    '''
    def __init__(self,
                 label: str = None,
                 payload: str = None,
                 color: str = KeyboardColor.SECONDARY):
        super().__init__(
            type = "text",
            label = label,
            payload = payload,
            color = color
        )

class LocationKeyboardButton(KeyboardButton):
    '''
    Кнопка, которая открывает окно с информацией о местоположении
    (см. https://dev.vk.ru/ru/api/bots/development/keyboard#location)
    '''
    def __init__(self,
                 payload: str = None):
        super().__init__(
            type = "location",
            payload = payload
        )

class VkPayKeyboardButton(KeyboardButton):
    '''
    Кнопка, которая открывает окно оплаты VK Pay с предопределенными параметрами
    (см. https://dev.vk.ru/ru/api/bots/development/keyboard#vkpay)
    '''
    def __init__(self,
                 hash: str = None,
                 payload: str = None):
        super().__init__(
            type = "vkpay",
            hash = hash,
            payload = payload
        )

class OpenLinkKeyboardButton(KeyboardButton):
    '''
    Кнопка, которая открывает окно оплаты VK Pay с предопределенными параметрами
    (см. https://dev.vk.ru/ru/api/bots/development/keyboard#open_link)
    '''
    def __init__(self,
                 link: str = None,
                 label: str = None,
                 payload: str = None):
        super().__init__(
            type = "open_link",
            link = link,
            label = label,
            payload = payload
        )

class OpenAppKeyboardButton(KeyboardButton):
    '''
    Кнопка, которая открывает окно оплаты VK Pay с предопределенными параметрами
    (см. https://dev.vk.ru/ru/api/bots/development/keyboard#open_app)
    '''
    def __init__(self,
                 app_id: int = None,
                 label: str = None,
                 owner_id: int = None,
                 hash: str = None,
                 payload: str = None):
        super().__init__(
            type = "open_app",
            app_id = app_id,
            label = label,
            owner_id = owner_id,
            hash = hash,
            payload = payload
        )

class CallbackActionKeyboardButton(KeyboardButton):
    '''
    Кнопка, которая открывает окно оплаты VK Pay с предопределенными параметрами
    (см. https://dev.vk.ru/ru/api/bots/development/keyboard#callback)
    '''
    def __init__(self,
                 label: str = None,
                 payload: str = None,
                 color: str = KeyboardColor.SECONDARY):
        super().__init__(
            type = "callback",
            label = label,
            payload = payload,
            color = color
        )


import pytest
import asyncio
import pytest_asyncio
import pytest_httpbin
import pytest_mock

import twilight_vk

class MockedHttpGet:

    def __init__(self):
        self.response = {"response": {"result": "ok"}}
        self.url = "https://api.example.com/"
        self.method = "mocked-method"
        self.status = 200
        self.headers = {'Content-Type': 'application/json'}
    
    async def json(self):
        return self.response
    
    async def get(self, *args, **kwargs):
        return self

async def fake_get_method(*args, **kwargs):
   _response = MockedHttpGet()
   return await _response.json()

@pytest.mark.asyncio
async def test_methods(monkeypatch, mocker):

    _client = twilight_vk.TwilightVK(ACCESS_TOKEN="123", loop_wrapper=asyncio.get_event_loop_policy())

    http_validate_mock = mocker.AsyncMock()
    http_validate_mock.side_effect = lambda response, *args, **kwargs: response
    mocker.patch('twilight_vk.framework.methods.base.HttpValidator.validate', http_validate_mock)

    monkeypatch.setattr(_client.__bot__.base_methods.httpClient, "get", MockedHttpGet().get)

    assert await _client.methods.messages.send() == {"response": {"result": "ok"}}
    assert await _client.methods.users.get(user_ids=1) == {"response": {"result": "ok"}}
    assert await _client.methods.groups.getById() == {"response": {"result": "ok"}}


from http import HTTPStatus

import pytest

from twilight_vk.utils.types.event_types import BotEventType
from twilight_vk.utils.types.response import Response

class MockedHttpResponse:

    def __init__(self,
                 *args,
                 **kwargs):
        self.status = kwargs.get('status') or HTTPStatus.OK
        self.method = kwargs.get('method') or "GET"
        self.url = "https://api.example.com/get"
        self.response = kwargs.get('response') or {
            "response": {
                "test": 1
            }
        }
        self.headers = kwargs.get('headers') or {"Content-Type": "application/json"}
    
    async def json(self):
        return self.response

class MockEvent:

    def __init__(self,
                 event_type: str):
        self.object = {
            "type": event_type or BotEventType.MESSAGE_NEW,
            "object": {
                "message": {
                    "conversation_message_id": 1,
                    "peer_id": 123,
                    "text": "test"
                }
            }
        }
    
    def get(self):
        return self.object

@pytest.fixture
def http_responses():
    return [
        MockedHttpResponse(
            status = HTTPStatus.OK,
        ),
        MockedHttpResponse(
            status = HTTPStatus.TEMPORARY_REDIRECT
        ),
        MockedHttpResponse(
            status = HTTPStatus.BAD_REQUEST
        )
    ]

@pytest.fixture
def event_responses():
    return [
        MockedHttpResponse(
            headers = {"Content-Type": "application/json"},
            response = {"response": {"test": 1}}
        ),
        MockedHttpResponse(
            headers = {'Invalid-header': 'Blep'},
        ),
        MockedHttpResponse(
            response = {"Invalid-response": "test"}
        ),
        MockedHttpResponse(
            response = {"ts": 0, "updates": []}
        ),
        MockedHttpResponse(
            response = {"failed": 3}
        ),
        MockedHttpResponse(
            response = {
                "error": {
                    "error_code": 1,
                    "error_msg": "Unknown error, try again later",
                    "request_params": [
                        {
                            "key": "v",
                            "value": "1.234"
                        }
                    ]
                }
            }
        )
    ]

@pytest.fixture
def event_responses_results():
    return [
        {"response": {"test": 1}},
        None,
        None,
        {"ts": 0, "updates": []},
        {"failed": 3},
        {
            "error": {
                "error_code": 1,
                "error_msg": "Unknown error, try again later",
                "request_params": [
                    {
                        "key": "v",
                        "value": "1.234"
                    }
                ]
            }
        }
    ]

@pytest.fixture
def request_params():
    return {
        "peer_ids": [123, 456],
        "extended": False,
        "forward": {"is_reply": True},
        "message": "abc-test"
    }

@pytest.fixture
def outputs():
    return [
        (MockEvent(event_type=BotEventType.MESSAGE_NEW).get(), "Test output"),
        (MockEvent(event_type=BotEventType.MESSAGE_NEW).get(), Response(
            peer_ids=[123],
            message="Test output"
        )),
        (MockEvent(event_type=BotEventType.LIKE_ADD).get(), "Test output"),
        (MockEvent(event_type=BotEventType.LIKE_ADD).get(), Response(
            peer_ids=[123],
            message="Test output"
        ))
    ]


from .buttons.buttons import (
    TextActionKeyboardButton,
    LocationKeyboardButton,
    VkPayKeyboardButton,
    OpenLinkKeyboardButton,
    OpenAppKeyboardButton,
    CallbackActionKeyboardButton
)
from .keyboard import KeyboardMarkup


import logging

from aiohttp import ClientResponse
from http import HTTPStatus

from ..exceptions.validator import HttpValidationError
from ...utils.validation_reports import make_it_colored as color

logger = logging.getLogger("http-validator")

class HttpValidator:

    async def _isValid(response: ClientResponse) -> bool:
        '''
        Response type checking
        '''
        logger.debug(f"HTTP Validation: Type matching check...")

        if type(response) != ClientResponse:
            logger.error(f"Types does not match: {type(response)} != {ClientResponse}")
            return False
        
        return True
    
    async def _isSuccess(response: ClientResponse) -> bool:
        '''
        Http status code check
        '''
        logger.debug(f"HTTP Validation: HTTP Status check...")

        if response.status >= HTTPStatus.BAD_REQUEST:
            logger.error(f"HTTP status code is {response.status}")
            return False
        
        return True


    async def validate(response: ClientResponse) -> ClientResponse:
        '''
        Validating http response
        '''
        logger.debug(f"Validating http response for [{response.method}] \"{response.url}\"...")

        isValid = await HttpValidator._isValid(response)
        isSuccess = None

        if isValid:
            isSuccess = await HttpValidator._isSuccess(response)

        if isValid and isSuccess:
            logger.debug(f"HTTP Response is valid")
            return response
        
        logger.warning(f"HTTP Response is not valid")
        logger.warning(f'{"Raw is valid":<15}|{"":>1}{"Is success":<15}|{"":>1}{"response":<10}')
        logger.warning(f'{color(isValid, indent=15)}|{"":>1}{color(isSuccess, indent=15)}|{"":>1}{color(response)}')
        
        raise HttpValidationError(isValid, isSuccess, response)


from ..logger.darky_visual import STYLE, FG, BG

def make_it_colored(value, color: str = None, indent: int = 10):

    _color = FG.CUSTOM_COLOR("#555")
    _text = "None"

    if isinstance(value, str):
        _color = FG.CUSTOM_COLOR("#888")
        _text = value

    if isinstance(value, bool):
        if value:
            _color = FG.GREEN
            _text = "YES"
        else:
            _color = FG.RED
            _text = "NO"
    
    if color:
        _color = color
    
    return f"{_color}{_text:<{indent}}{STYLE.RESET}"


import json

import pytest
import pytest_asyncio

from twilight_vk.framework.validators import (
    HttpValidator,
    EventValidator,
    RequestValidator,
    ResponseValidator
)
from twilight_vk.framework.exceptions import(
    HttpValidationError,
    EventValidationError,
    VkApiError,
    LongPollError
)
from twilight_vk.utils.types.response import Response
from tests.fixtures.validators import (
    MockedHttpResponse,
    MockEvent,
    http_responses,
    event_responses,
    event_responses_results,
    request_params,
    outputs
)

async def fake_type_validating(*args, **kwargs):
    return True

@pytest.mark.asyncio
async def test_http_validator(http_responses, monkeypatch):
    monkeypatch.setattr(HttpValidator, "_isValid", fake_type_validating)

    # HTTPStatus.OK
    assert await HttpValidator.validate(http_responses[0]) == http_responses[0]

    # HTTPStatus.TEMPORARY_REDIRECT
    assert await HttpValidator.validate(http_responses[1]) == http_responses[1]

    # HTTPStatus.BAD_REQUEST
    with pytest.raises(HttpValidationError, match="Response validation error: Status code is not success "):
        await HttpValidator.validate(http_responses[2])

@pytest.mark.asyncio
async def test_event_validator(event_responses, event_responses_results):

    # Valid response
    assert await EventValidator.validate(event_responses[0]) == event_responses_results[0]

    # Invalid header
    with pytest.raises(EventValidationError, match="Response validation error: Content is not JSON "):
        await EventValidator.validate(event_responses[1])

    # Invalid response
    with pytest.raises(EventValidationError, match="Response validation error: Response doesn't contain the required fields "):
        await EventValidator.validate(event_responses[2])
    
    # Response from polling
    assert await EventValidator.validate(event_responses[3], True) == event_responses_results[3]
    
    # Error while polling
    with pytest.raises(LongPollError):
        await EventValidator.validate(event_responses[4])
    
    # Error while requesting VK API
    with pytest.raises(VkApiError):
        await EventValidator.validate(event_responses[5])

@pytest.mark.asyncio
async def test_request_validator(request_params):
    assert await RequestValidator.validate(request_params) == {
        "peer_ids": "123,456",
        "extended": "false",
        "forward": "{\"is_reply\": true}",
        "message": "abc-test"
    }

@pytest.mark.asyncio
async def test_response_validator(outputs):

    # MESSAGE_NEW with string output
    assert (await ResponseValidator.validate(outputs[0][1], 
                                             outputs[0][0])).getData() == Response(peer_ids=123,
                                                                                   message="Test output",
                                                                                   forward={"peer_id": 123,
                                                                                            "conversation_message_ids": 1,
                                                                                            "is_reply": True}).getData()

    # MESSAGE_NEW with Response() output
    assert (await ResponseValidator.validate(outputs[1][1], 
                                             outputs[1][0])).getData() == Response(peer_ids=[123], 
                                                                                   message="Test output").getData()

    # LIKE_ADD with string output
    assert (await ResponseValidator.validate(outputs[2][1], 
                                             outputs[2][0])) == "Test output"

    # LIKE_ADD with Response() output
    assert (await ResponseValidator.validate(outputs[3][1], 
                                             outputs[3][0])).getData() == Response(peer_ids=[123], 
                                                                                   message="Test output").getData()


from .base import (
    BASE_EVENT_HANDLER
)

class DEFAULT_HANDLER(BASE_EVENT_HANDLER):
    pass
    

class MESSAGE_NEW(DEFAULT_HANDLER):
    pass

class MESSAGE_REPLY(DEFAULT_HANDLER):
    pass

class MESSAGE_EDIT(DEFAULT_HANDLER):
    pass

class MESSAGE_TYPING_STATE(DEFAULT_HANDLER):
    pass

class MESSAGE_READ(DEFAULT_HANDLER):
    pass

class MESSAGE_EVENT(DEFAULT_HANDLER):
    pass
    

class PHOTO_NEW(DEFAULT_HANDLER):
    pass

class PHOTO_COMMENT_NEW(DEFAULT_HANDLER):
    pass

class PHOTO_COMMENT_EDIT(DEFAULT_HANDLER):
    pass

class PHOTO_COMMENT_RESTORE(DEFAULT_HANDLER):
    pass

class PHOTO_COMMENT_DELETE(DEFAULT_HANDLER):
    pass


class AUDIO_NEW(DEFAULT_HANDLER):
    pass


class VIDEO_NEW(DEFAULT_HANDLER):
    pass

class VIDEO_COMMENT_NEW(DEFAULT_HANDLER):
    pass

class VIDEO_COMMENT_EDIT(DEFAULT_HANDLER):
    pass

class VIDEO_COMMENT_RESTORE(DEFAULT_HANDLER):
    pass

class VIDEO_COMMENT_DELETE(DEFAULT_HANDLER):
    pass


class WALL_POST_NEW(DEFAULT_HANDLER):
    pass

class WALL_REPOST(DEFAULT_HANDLER):
    pass

class WALL_SCHEDULE_POST_NEW(DEFAULT_HANDLER):
    pass

class WALL_SCHEDULE_POST_DELETE(DEFAULT_HANDLER):
    pass

class WALL_REPLY_NEW(DEFAULT_HANDLER):
    pass

class WALL_REPLY_EDIT(DEFAULT_HANDLER):
    pass

class WALL_REPLY_RESTORE(DEFAULT_HANDLER):
    pass

class WALL_REPLY_DELETE(DEFAULT_HANDLER):
    pass


class LIKE_ADD(DEFAULT_HANDLER):
    pass

class LIKE_REMOVE(DEFAULT_HANDLER):
    pass


class MARKET_COMMENT_NEW(DEFAULT_HANDLER):
    pass

class MARKET_COMMENT_EDIT(DEFAULT_HANDLER):
    pass

class MARKET_COMMENT_RESTORE(DEFAULT_HANDLER):
    pass

class MARKET_COMMENT_DELETE(DEFAULT_HANDLER):
    pass

class MARKET_ORDER_NEW(DEFAULT_HANDLER):
    pass

class MARKET_ORDER_EDIT(DEFAULT_HANDLER):
    pass


class GROUP_LEAVE(DEFAULT_HANDLER):
    pass

class GROUP_JOIN(DEFAULT_HANDLER):
    pass

class USER_BLOCK(DEFAULT_HANDLER):
    pass

class USER_UNBLOCK(DEFAULT_HANDLER):
    pass


class POLL_VOTE_NEW(DEFAULT_HANDLER):
    pass

class GROUP_OFFICERS_EDIT(DEFAULT_HANDLER):
    pass

class GROUP_CHANGE_SETTINGS(DEFAULT_HANDLER):
    pass

class GROUP_CHANGE_PHOTO(DEFAULT_HANDLER):
    pass

class VKPAY_TRANSACTION(DEFAULT_HANDLER):
    pass


import logging

from .buttons.buttons import KeyboardButton

logger = logging.getLogger("keyboard")

class KeyboardMarkup:
    '''
    Клавиатура ВК
    '''
    def __init__(self,
                 one_time: bool = False,
                 inline: bool = False,
                 buttons: list[list[KeyboardButton]] = []):
        
        self._one_time: bool = None
        self._inline: bool = inline
        
        if not self._inline and one_time:
            self._one_time: bool = one_time

        self._buttons: list[list[KeyboardButton]] = buttons

        logger.debug("Generating keyboard markup...")

        self._buttons_markup: list[list[KeyboardButton]] = []

        for _line in self._buttons:
            self._buttons_markup.append([])
            for _button in _line:
                self._buttons_markup[-1].append(_button.getButton())

        self._keyboard_markup: dict = {
            "inline": self._inline,
            "buttons": self._buttons_markup
        }

        if not self._inline:
            self._keyboard_markup.setdefault("one_time", self._one_time)
        logger.debug(f"Keyboard markup was successfully generated: {self._keyboard_markup}")
    
    def getMarkup(self):
        '''
        Возвращает рамзетку клавиатуры
        '''
        return self._keyboard_markup


import logging

from ...types.keyboard_colors import KeyboardColor

logger = logging.getLogger('keyboard')

class KeyboardButton:
    '''
    Keyboard button
    '''
    def __init__(self,
                 type: str = None,
                 label: str = None,
                 payload: str = None,
                 hash: str = None,
                 link: str = None,
                 app_id: int = None,
                 owner_id: int = None,
                 color: str = KeyboardColor.SECONDARY):
        '''
        Базовый класс кнопки для клавиатуры, поддерживающий все параметры для универсальной конфигурации
        (см. https://dev.vk.ru/ru/api/bots/development/keyboard#Типы%20кнопок)
        '''
        self._type: str = type

        self._label: str = None
        self._payload: str = None

        if label:
            if len(label) > 40:
                logger.warning(f"The maximum length of the label is 40 characters. Label will be cut off to the limits")
                self._label: str = f"{label[:37]}..."
            self._label: str = f"{label}"

        if payload:
            if len(payload) > 255:
                logger.error(f"The maximum length of the payload is 255 characters.")
                raise ValueError("payload's length should be less than 255 characters.")
            self._payload: str = payload

        self._hash: str = hash
        self._link: str = link
        self._app_id: int = app_id
        self._owner_id: int = owner_id

        self._color: str = color

        logger.debug(f"Generating keyboard button's markup with type: {self._type}")
        
        self._button = {
            "action": {
                "type": self._type
            }
        }

        if self._type in ["text", "callback"]:
            self._button.setdefault("color", self._color)

        if self._type in ["text", "callback", "open_link", "open_app"]:
            self._button["action"].setdefault("label", self._label)
        
        if self._payload:
            self._button["action"].setdefault("payload", self._payload)
        
        if (self._type == "vkpay") or (self._type == "open_app" and self._hash):
            self._button["action"].setdefault("hash", self._hash)

        if self._type == "open_link":
            self._button["action"].setdefault("link", self._link)
        
        if self._type == "open_app":
            self._button["action"].setdefault("app_id", self._app_id)
            if self._owner_id:
                self._button["action"].setdefault("owner_id", self._owner_id)
        
        logger.debug(f"Button was successfully generated: {self._button}")

    def getButton(self):
        '''
        Возвращает JSON описывающий кнопку
        '''
        return self._button



from typing import TYPE_CHECKING
import logging

from ...utils.types.event_types import BotEventType
from ...utils.types.response import Response

logger = logging.getLogger("resp-validator")

class ResponseValidator:

    async def validate(response: 'Response | str',
                       event: dict = None):
        '''
        Response validator for custom handlers function's output

        :param response: Function's output
        :type response: str | Response

        :param event: Event object
        :type event: dict
        '''

        _event_type = event.get("type")
        _response_body = response

        logger.debug(f"Validating response for event {_event_type} with type {type(_response_body)}...")

        if _event_type in [
            BotEventType.MESSAGE_NEW,
            BotEventType.MESSAGE_REPLY,
            BotEventType.MESSAGE_EDIT,
            BotEventType.MESSAGE_TYPING_STATE,
            BotEventType.MESSAGE_READ,
            BotEventType.MESSAGE_EVENT
        ]:
            if isinstance(response, str):
                logger.debug(f"Got <string> output, converting to the <Response>...")
                _response_body = Response(
                    peer_ids=event["object"]["message"]["peer_id"],
                    message=f"{response}",
                    forward={
                        "peer_id": event["object"]["message"]["peer_id"],
                        "conversation_message_ids": event["object"]["message"]["conversation_message_id"],
                        "is_reply": True
                    }
                )
        
        logger.debug(f"Response body has been validated")
        
        return _response_body



import pytest
import pytest_asyncio
import pytest_httpbin

from aiohttp import ClientResponse
from http import HTTPStatus

from tests.fixtures.http import (
    http_client,
    http_params,
    http_data,
    Http
)

@pytest.mark.asyncio
async def test_init_session(http_client: Http,
                            httpbin):
    assert http_client.headers == {"X-Test": "test"}
    assert http_client.timeout == 10
    assert http_client.session is None

@pytest.mark.asyncio
async def test_http_get_raw(http_client: Http,
                            http_params: dict,
                            httpbin):
    response = await http_client.get(
        url = f"{httpbin.url}/get",
        params = http_params,
        raw = True
    )
    assert isinstance(response, ClientResponse)
    assert response.status == HTTPStatus.OK
    _json_response: dict = await response.json()
    assert _json_response.get("args") == http_params
    

@pytest.mark.asyncio
async def test_http_get_json(http_client: Http,
                             http_params: dict,
                             httpbin):
    response = await http_client.get(
        url = f"{httpbin.url}/get",
        params = http_params,
        raw = False
    )
    assert isinstance(response, dict)
    assert response.get("args") == http_params
    assert response.get("headers").get("X-Test") == "test"

@pytest.mark.asyncio
async def test_http_post_raw(http_client: Http,
                             http_params: dict,
                             http_data: dict,
                             httpbin):
    response = await http_client.post(
        url = f"{httpbin.url}/post",
        data = http_data,
        params = http_params,
        raw = True
    )
    assert isinstance(response, ClientResponse)
    assert response.status == HTTPStatus.OK
    _json_response: dict = await response.json()
    assert _json_response.get("json") == http_data
    assert _json_response.get("args") == http_params

@pytest.mark.asyncio
async def test_http_post_json(http_client: Http,
                             http_params: dict,
                             http_data: dict,
                             httpbin):
    response = await http_client.post(
        url = f"{httpbin.url}/post",
        data = http_data,
        params = http_params,
        raw = False
    )
    assert isinstance(response, dict)
    assert response["json"] == http_data
    assert response["args"] == http_params
    assert response["headers"]["X-Test"] == "test"

@pytest.mark.asyncio
async def test_close_session(http_client: Http):
    await http_client._get_session()
    assert http_client is not None
    assert not http_client.session.closed
    await http_client.close()
    assert http_client.session is None


import pytest

@pytest.fixture
def inline_keyboard_markup():
    return {
        "inline": True,
        "buttons": [
            [
                {
                    "action": {
                        "type": "text",
                        "label": "TestButton"
                    },
                    "color": "positive"
                }
            ]
        ]
    }

@pytest.fixture
def keyboard_markup():
    return {
        "one_time": True,
        "inline": False,
        "buttons": [
            [
                {
                    "action": {
                        "type": "text",
                        "payload": "{\"button\": \"1\"}",
                        "label": "TestButton"
                    },
                    "color": "secondary"
                },
                {
                    "action":{
                        "type":"location",
                        "payload":"{\"button\": \"2\"}"
                    }
                }
            ],
            [
                {
                    "action":{
                        "type":"vkpay",
                        "hash":"action=transfer-to-group&group_id=12345&aid=10"
                    }
                }
            ]
        ]
    }


from typing import TYPE_CHECKING
import logging
import asyncio

from .event_handlers import *
from ...utils.types.event_types import BotEventType

if TYPE_CHECKING:
    from ..methods import VkMethods
    from ...utils.event_loop import TwiTaskManager

class OnEventLabeler:

    def __init__(self, handlers: dict):
        self.handlers: dict[str, BASE_EVENT_HANDLER] = handlers

    def raw(self, event_type: str, *rules):
        def decorator(func):
            self.handlers[event_type].add(func, rules)
            return func
        return decorator

    def all(self, *rules):
        def decorator(func):
            for handler_name in self.handlers.keys():
                self.handlers[handler_name].add(func, rules)
            return func
        return decorator
    
    def message_new(self, *rules):
        def decorator(func):
            self.handlers[BotEventType.MESSAGE_NEW].add(func, rules)
            return func
        return decorator

class EventRouter:

    def __init__(self,
                 vk_methods:'VkMethods',
                 _loop_wrapper: 'TwiTaskManager'):
        '''
        Router for events.
        Allows to route events to separate event handlers
        '''
        self.vk_methods = vk_methods
        self._loop_wrapper = _loop_wrapper
        self.logger = logging.getLogger("event-router")
        self._handlers = {
            "default": DEFAULT_HANDLER(self.vk_methods),

            # Message events
            BotEventType.MESSAGE_NEW: MESSAGE_NEW(self.vk_methods),
            BotEventType.MESSAGE_REPLY: MESSAGE_REPLY(self.vk_methods),
            BotEventType.MESSAGE_EDIT: MESSAGE_EDIT(self.vk_methods),
            BotEventType.MESSAGE_TYPING_STATE: MESSAGE_TYPING_STATE(self.vk_methods),
            BotEventType.MESSAGE_READ: MESSAGE_READ(self.vk_methods),
            BotEventType.MESSAGE_EVENT: MESSAGE_EVENT(self.vk_methods),
            
            # Photo events
            BotEventType.PHOTO_NEW: PHOTO_NEW(self.vk_methods),
            BotEventType.PHOTO_COMMENT_NEW: PHOTO_COMMENT_NEW(self.vk_methods),
            BotEventType.PHOTO_COMMENT_EDIT: PHOTO_COMMENT_EDIT(self.vk_methods),
            BotEventType.PHOTO_COMMENT_RESTORE: PHOTO_COMMENT_RESTORE(self.vk_methods),
            BotEventType.PHOTO_COMMENT_DELETE: PHOTO_COMMENT_DELETE(self.vk_methods),
            
            # Audio events
            BotEventType.AUDIO_NEW: AUDIO_NEW(self.vk_methods),
            
            # Video events
            BotEventType.VIDEO_NEW: VIDEO_NEW(self.vk_methods),
            BotEventType.VIDEO_COMMENT_NEW: VIDEO_COMMENT_NEW(self.vk_methods),
            BotEventType.VIDEO_COMMENT_EDIT: VIDEO_COMMENT_EDIT(self.vk_methods),
            BotEventType.VIDEO_COMMENT_RESTORE: VIDEO_COMMENT_RESTORE(self.vk_methods),
            BotEventType.VIDEO_COMMENT_DELETE: VIDEO_COMMENT_DELETE(self.vk_methods),
            
            # Wall events
            BotEventType.WALL_POST_NEW: WALL_POST_NEW(self.vk_methods),
            BotEventType.WALL_REPOST: WALL_REPOST(self.vk_methods),
            BotEventType.WALL_SCHEDULE_POST_NEW: WALL_SCHEDULE_POST_NEW(self.vk_methods),
            BotEventType.WALL_SCHEDULE_POST_DELETE: WALL_SCHEDULE_POST_DELETE(self.vk_methods),
            BotEventType.WALL_REPLY_NEW: WALL_REPLY_NEW(self.vk_methods),
            BotEventType.WALL_REPLY_EDIT: WALL_REPLY_EDIT(self.vk_methods),
            BotEventType.WALL_REPLY_RESTORE: WALL_REPLY_RESTORE(self.vk_methods),
            BotEventType.WALL_REPLY_DELETE: WALL_REPLY_DELETE(self.vk_methods),
            
            # Like events
            BotEventType.LIKE_ADD: LIKE_ADD(self.vk_methods),
            BotEventType.LIKE_REMOVE: LIKE_REMOVE(self.vk_methods),
            
            # Board events
            # BotEventType.BOARD_POST_NEW: BOARD_POST_NEW(self.vk_methods),
            # BotEventType.BOARD_POST_EDIT: BOARD_POST_EDIT(self.vk_methods),
            # BotEventType.BOARD_POST_RESTORE: BOARD_POST_RESTORE(self.vk_methods),
            # BotEventType.BOARD_POST_DELETE: BOARD_POST_DELETE(self.vk_methods),
            
            # Market events
            # BotEventType.MARKET_COMMENT_NEW: MARKET_COMMENT_NEW(self.vk_methods),
            # BotEventType.MARKET_COMMENT_EDIT: MARKET_COMMENT_EDIT(self.vk_methods),
            # BotEventType.MARKET_COMMENT_RESTORE: MARKET_COMMENT_RESTORE(self.vk_methods),
            # BotEventType.MARKET_COMMENT_DELETE: MARKET_COMMENT_DELETE(self.vk_methods),
            # BotEventType.MARKET_ORDER_NEW: MARKET_ORDER_NEW(self.vk_methods),
            # BotEventType.MARKET_ORDER_EDIT: MARKET_ORDER_EDIT(self.vk_methods),
            
            # Group events
            BotEventType.GROUP_LEAVE: GROUP_LEAVE(self.vk_methods),
            BotEventType.GROUP_JOIN: GROUP_JOIN(self.vk_methods),
            BotEventType.USER_BLOCK: USER_BLOCK(self.vk_methods),
            BotEventType.USER_UNBLOCK: USER_UNBLOCK(self.vk_methods),
            
            # Other events
            # BotEventType.POLL_VOTE_NEW: POLL_VOTE_NEW(self.vk_methods),
            # BotEventType.GROUP_OFFICERS_EDIT: GROUP_OFFICERS_EDIT(self.vk_methods),
            # BotEventType.GROUP_CHANGE_SETTINGS: GROUP_CHANGE_SETTINGS(self.vk_methods),
            # BotEventType.GROUP_CHANGE_PHOTO: GROUP_CHANGE_PHOTO(self.vk_methods),
            # BotEventType.VKPAY_TRANSACTION: VKPAY_TRANSACTION(self.vk_methods),
        }
        self.on_event = OnEventLabeler(self._handlers)
    
    async def handle(self, polling_response: dict):
        '''
        Handles the event list

        :param polling_response: Response from the polling requests contains "ts" and "updates" keys
        :type polling_response: dict
        '''
        self.logger.debug("Routing the events...")
        events = []
        for event in polling_response["updates"]:
            events.append(
                self._loop_wrapper._loop.create_task(
                    self.route(event)
                )
            )
        try:
            await asyncio.gather(*events, return_exceptions=False)
        except Exception as exc:
            self.logger.error(f"{exc.__class__.__name__}: {exc}", exc_info=True)

    async def route(self, current_event: dict):
        '''
        Routing the event to the exact handler
        '''
        event_type = current_event.get("type", "default")
        handler: BASE_EVENT_HANDLER = self._handlers.get(event_type, self._handlers["default"])
        self.logger.debug(f"Routing the event {current_event.get("type")} to the {handler.__class__.__name__} handler")
        await handler.execute(current_event)


import pytest
import pytest_asyncio

from twilight_vk.framework.handlers.event_handlers import DEFAULT_HANDLER
from tests.fixtures.rules import (
    BaseRule,
    fake_event,
    messages_list,
    results,
    handler_results,
    rules_list,
    MockVkMethods
)

@pytest.mark.asyncio
async def test_rules(fake_event: dict, messages_list: list, results: list, rules_list: list[BaseRule]):
    for test in range(len(messages_list["messages"])):
        fake_event["object"]["message"]["text"] = messages_list["messages"][test]
        fake_event["object"]["message"]["reply_message"] = messages_list["replies"][test]
        if fake_event["object"]["message"]["reply_message"] == None:
            fake_event["object"]["message"].__delitem__("reply_message")
        fake_event["object"]["message"]["fwd_messages"] = messages_list["fwds"][test]
        fake_event["object"]["message"]["action"] = messages_list["actions"][test]
        if fake_event["object"]["message"]["action"] == None:
            fake_event["object"]["message"].__delitem__("action")
        rule_results = [await rule._check(fake_event) for rule in rules_list]
        assert rule_results == results[test]

@pytest.mark.asyncio
async def test_handlerinput(fake_event: dict, messages_list: list, results: list, rules_list: list[BaseRule], handler_results: list):
    for test in range(len(messages_list["messages"])):
        fake_event["object"]["message"]["text"] = messages_list["messages"][test]
        rule_results = [await rule._check(fake_event) for rule in rules_list]
        handler_result = await DEFAULT_HANDLER(MockVkMethods())._extractArgs(rule_results)
        assert handler_result == handler_results[test]


import asyncio

import pytest
import pytest_asyncio
from aiohttp import ClientResponse

from twilight_vk import TwilightVK

@pytest_asyncio.fixture
def bot(_function_event_loop):
    _bot = TwilightVK(
        ACCESS_TOKEN="123",
        GROUP_ID=123,
        loop_wrapper=_function_event_loop
    )
    yield _bot

class MockEventGenerator:

    def __init__(self,
                 event_type: str = "default"):
        self.method = "GET"
        self.url = "https://api.example.com/"
        self.headers = {"Content-Type": "application/json"}
        self.body = {}

        match event_type:
            
            case "default":
                self.body = {
                    "ts": 123,
                    "updates": []
                }
        
            case "auth success":
                self.body = {
                    "response": {
                        "server": self.url,
                        "key": "fakekey123",
                        "ts": 123
                    }
                }
        
            case "auth fail":
                self.body = {
                    "error": {
                        "error_code": 5,
                        "error_msg": "User authorization failed: invalid access_token (4).",
                        "request_params": [
                            {
                                "key": "access_token",
                                "value": "123"
                            },
                            {
                                "key": "group_id",
                                "value": 123
                            },
                            {
                                "key": "v",
                                "value": "1.234"
                            }
                        ]
                    }
                }
            
            case "message_typing_state":
                self.body = {
                    "ts": 123,
                    "updates": [
                        {
                            "group_id": 123,
                            "type": "message_typing_state",
                            "object": {
                                "from_id": 123,
                                "to_id": -123,
                                "state": "typing"
                            }
                        }
                    ]
                }

            case "message_new":
                self.body = {
                    "ts": 123,
                    "updates": [
                        {
                            "group_id": 123,
                            "type": "message_new",
                            "object": {
                                "client_info": {},
                                "message": {
                                    "from_id": 123,
                                    "id": 1,
                                    "conversation_message_id": 1,
                                    "text": "a",
                                    "peer_id": 2000000001
                                }
                            }
                        }
                    ]
                }
            
            case "like_add":
                self.body = {
                    "ts": 123,
                    "updates": [
                        {
                            "group_id": 123,
                            "type": "like_add",
                            "object": {
                                "liker_id": 123,
                                "object_type": "post",
                                "object_owner_id": 123,
                                "object_id": 1,
                                "thread_reply_id": 1,
                                "post_id": 1
                            }
                        }
                    ]
                }
            
            case "failed event 1":
                self.body = {
                    "failed": 1,
                    "ts": 321
                }
            
            case "failed event 2":
                self.body = {
                    "failed": 2
                }

            case "failed event 3":
                self.body = {
                    "failed": 3
                }
        
    async def json(self):
        return self.body
    
    async def get(self, *args, **kwargs):
        return self


import pytest
import pytest_asyncio

from twilight_vk import TwilightVK
from twilight_vk.utils.types.event_types import BotEventType

@pytest_asyncio.fixture
def bot(_function_event_loop):
    _bot = TwilightVK(
        ACCESS_TOKEN="123",
        GROUP_ID=123,
        loop_wrapper=_function_event_loop
    )
    yield _bot


@pytest.fixture
def response():
    return {
        "ts": 123,
        "updates": [
            {
                "type": BotEventType.MESSAGE_TYPING_STATE,
                "object": {
                    "state": "typing",
                    "from_id": 123,
                    "to_id": -123
                }
            },
            {
                "type": BotEventType.MESSAGE_NEW,
                "object": {
                    "client_info": {},
                    "message": {
                        "conversation_message_id": 1,
                        "peer_id": 123,
                        "text": "Test"
                    }
                }
            },
            {
                "type": BotEventType.LIKE_ADD,
                "object": {
                    "liker_id": 123,
                    "object_type": "post",
                    "object_owner_id": 123,
                    "object_id": 1,
                    "thread_reply_id": 1,
                    "post_id": 1
                }
            }
        ]
    }